import unittest
try:
    from unittest.mock import patch, MagicMock
except ImportError:
    from mock import patch, MagicMock
from acky.aws import AWS
from acky.api import AwsApiClient
import acky.s3
import botocore.session


class TestAWS(unittest.TestCase):
    @patch('botocore.session.get_session')
    def test_gets_session(self, _get_session):
        AWS('region', profile='profile')
        _get_session.assert_called()

    @patch('botocore.session.get_session')
    @patch('acky.iam.UserCollection.get_current')
    def test_gets_userinfo(self, _get_current_user, _get_session):
        aws = AWS('region')
        aws.userinfo
        _get_current_user.assert_called()


class TestAwsApiClient(unittest.TestCase):
    class _HelpApiClient(AwsApiClient):
        service_name = "support"

    @patch('botocore.service.get_service')
    def test_gets_service(self, _get_service):
        self._HelpApiClient(AWS('region'))
        _get_service.assert_called()

    @patch('botocore.operation.Operation.call')
    @patch('botocore.service.Service.get_operation')
    def test_call(self, get_operation, call):
        self._HelpApiClient(AWS('region'))
        get_operation.assert_called()
        call.assert_called()


class TestS3(unittest.TestCase):
    """Mock tests for S3. Ensure botocore calls are correct and happening."""

    class _AWS(object):
        """AWS object for mock testing with only basic features."""
        def __init__(self):
            self.session = botocore.session.get_session()
            self.region = 'us-east-1'

        @property
        def s3(self):
            return acky.s3.S3(self)

    acky.aws.AWS = MagicMock(return_value=_AWS())
    aws = acky.aws.AWS()

    @patch('acky.s3.S3.call')
    @patch('botocore.operation.Operation.call')
    def test_get(self, boto_call, acky_call):
        self.aws.s3.get("s3://test/")
        acky_call.assert_called_with("ListObjects",
                                     response_data_key="Contents",
                                     Bucket="test", Delimiter="/")
        boto_call.assert_called()
        self.aws.s3.get("s3://")
        acky_call.assert_called_with("ListBuckets",
                                     response_data_key="Buckets")
        boto_call.assert_called()

    @patch('acky.s3.S3.call')
    @patch('botocore.operation.Operation.call')
    def test_create(self, boto_call, acky_call):
        self.aws.s3.create("s3://test/")
        acky_call.assert_called_with("CreateBucket", bucket="test")
        boto_call.assert_called()

    @patch('acky.s3.S3.call')
    @patch('botocore.operation.Operation.call')
    def test_destroy(self, boto_call, acky_call):
        self.aws.s3.destroy("s3://test/")
        acky_call.assert_called_with("DeleteBucket", bucket="test")
        boto_call.assert_called()

    @patch('acky.s3.S3.call')
    @patch('botocore.operation.Operation.call')
    def test_download(self, boto_call, acky_call):
        try:
            self.aws.s3.download("s3://test/a", "/tmp/test")
        except TypeError:
            # We can't actually write files with mocks.
            pass
        acky_call.assert_called_with("GetObject", bucket="test", key='a')
        boto_call.assert_called()

    @patch('acky.s3.S3.call')
    @patch('botocore.operation.Operation.call')
    def test_copy(self, boto_call, acky_call):
        self.aws.s3.copy("s3://test/", "s3://test2/")
        acky_call.assert_called_with("CopyObject", copy_source="test/",
                                     key="", bucket="test2")
        boto_call.assert_called()

    @patch('acky.s3.S3.copy')
    @patch('acky.s3.S3.destroy')
    def test_move(self, destroy, copy):
        self.aws.s3.move("s3://test/a", "s3://test/b")
        copy.assert_called_with("s3://test/a", "s3://test/b")
        destroy.assert_called_with("s3://test/a")


if __name__ == '__main__':
    unittest.main()



import unittest
import os
from pprint import pprint
import acky.aws
from acky.s3 import _parse_url, InvalidURL
import random

AWS = acky.aws.AWS('us-east-1', 'sbx')
TEST_BUCKET = "ackytest1"
TEST_PREFIX = "ackytest_"
BAD_URLS = ("", "s3:/", "s3:///", "s3:////file", "s3:////path/",
            "bucket/file", "bucket/dir/",
            "bucket/", "bucket")


def functional_tests_enabled():
    """Only do functional tests if the 'FUNCTIONAL' environment variable is
    set to 'yes'.
    """
    return os.environ.get('FUNCTIONAL', False) == 'yes'


def _get_values(arr, key='Key'):
    return [e.get(key, None) for e in arr]


class TestS3Functional(unittest.TestCase):
    """Functional tests for all S3 methods. Testing get() requires the
    TEST_BUCKET to be properly set up.
    TODO: automate setting up TEST_BUCKET.
    """

    def test_parse_url(self):
        test_urls = ["s3://",
                     "s3://bucket",
                     "s3://bucket/",
                     "s3://bucket/key",
                     "s3://bucket/dir/",
                     "s3://bucket/dir/key"]
        desired_parts = [('', ''),
                         ('bucket', ''),
                         ('bucket', ''),
                         ('bucket', 'key'),
                         ('bucket', 'dir/'),
                         ('bucket', 'dir/key')]
        for url, parts in zip(test_urls, desired_parts):
            assert _parse_url(url) == parts, "URL:  {0} improperly parsed."
        # Invalid URLS
        for bad_url in BAD_URLS:
            try:
                AWS.s3.get(bad_url)
                assert False, ("Invalid URL: '{0}' "
                               "did not trigger exception.".format(bad_url))
            except InvalidURL:
                pass

    @unittest.expectedFailure
    @unittest.skipUnless(functional_tests_enabled(),
                         "Functional tests not enabled.")
    def test_get(self):
        # Fetching lists of buckets
        buckets = [AWS.s3.get(), AWS.s3.get("s3://")]
        assert buckets[0] == buckets[1],\
            "Inconsistent operation for s3.get() when fetching buckets."
        pprint("Existing bucket names: {0}".format(_get_values(buckets[0],
                                                               'Name')))

        # Fetching buckets contents
        bucket_contents = [AWS.s3.get("s3://{0}".format(TEST_BUCKET)),
                           AWS.s3.get("s3://{0}/".format(TEST_BUCKET))]
        assert buckets[0] == buckets[1],\
            "Inconsistent operation for s3.get "\
            "when fetching buckets contents."
        pprint("Bucket keys: {0}".format(_get_values(bucket_contents[0])))

        # Fetching individual objects
        bucket_file = AWS.s3.get("s3://{0}/file".format(TEST_BUCKET))
        assert bucket_file['Key'] == 'file',\
            "Incorrect operation for s3.get "\
            "when fetching file."

        # Fetching directory trees
        bucket_contents = [AWS.s3.get("s3://{0}/dir".format(TEST_BUCKET)),
                           AWS.s3.get("s3://{0}/dir/".format(TEST_BUCKET))]
        assert bucket_contents[0] == bucket_contents[1],\
            "Inconsistent operation for s3.get "\
            "when fetching directory contents."
        # assert _get_values(buckets[0]) == [test_values, ...],\
        #     "Incorrect objects fetched by s3.get "\
        #     "when fetching directory trees."

    @unittest.skipUnless(functional_tests_enabled(),
                         "Functional tests not enabled.")
    def test_create_destroy(self):
        # Test s3.create()
        AWS.s3.create("s3://{0}create_test".format(TEST_PREFIX))
        AWS.s3.create("s3://{0}create_test/create_test".format(TEST_PREFIX))
        AWS.s3.create("s3://{0}create_test/"
                      "create_test/test".format(TEST_PREFIX))
        buckets = AWS.s3.get()
        assert "{0}create_test".format(TEST_PREFIX) in\
            _get_values(buckets, 'Name'), "Could not create bucket."
        assert AWS.s3.get("s3://{0}create_test/"
                          "create_test".format(TEST_PREFIX)) and\
            AWS.s3.get("s3://{0}create_test/"
                       "create_test/test".format(TEST_PREFIX)),\
            "Could not create objects in {0}create_test".format(TEST_PREFIX)

        # Test s3.destroy()
        # File deletion
        AWS.s3.destroy("s3://{0}create_test/create_test".format(TEST_PREFIX),
                       recursive=False)
        bucket_objects = AWS.s3.get("s3://{0}create_test".format(TEST_PREFIX))
        assert 'create_test' not in _get_values(bucket_objects),\
               "Inaccurate destroy() method."

        # Recursive directory deletion
        AWS.s3.destroy("s3://{0}create_test/create_test/".format(TEST_PREFIX),
                       recursive=True)
        bucket_objects = AWS.s3.get("s3://{0}create_test/"
                                    "create_test/".format(TEST_PREFIX))
        assert 'create_test/test' not in _get_values(bucket_objects),\
               "Inaccurate deletion in destroy() with recursive=True."

        # Recursive bucket deletion
        AWS.s3.destroy("s3://{0}create_test".format(TEST_PREFIX),
                       recursive=True)
        buckets = AWS.s3.get()
        assert not "{0}create_test".format(TEST_PREFIX) in _get_values(buckets,
                                                                       'Name')

    @unittest.skipUnless(functional_tests_enabled(),
                         "Functional tests not enabled.")
    def test_upload_download(self):
        test_value = str(random.randint(100000000, 999999999))
        with open("/tmp/acky_test", 'w') as temp_file:
            temp_file.write(test_value)
        AWS.s3.create("s3://{0}upload_test".format(TEST_PREFIX))
        AWS.s3.upload("/tmp/acky_test",
                      "s3://{0}upload_test/upload".format(TEST_PREFIX))
        bucket_objects = AWS.s3.get("s3://{0}upload_test".format(TEST_PREFIX))
        assert len(bucket_objects) == 1, "Testing bucket unclean."
        assert 9 == bucket_objects[0]['Size'],\
            "Data upload error: size does not match."
        AWS.s3.download("s3://{0}upload_test/upload".format(TEST_PREFIX),
                        "/tmp/acky_test_dl")

        with open("/tmp/acky_test_dl", 'r') as temp_file:
            assert temp_file.read() == test_value,\
                "Download error: incorrect data."

if __name__ == "__main__":
    unittest.main()



from acky.api import AwsApiClient
try:
    from urllib import parse
except ImportError:
    import urlparse as parse


class InvalidURL(Exception):
    def __init__(self, url, msg=None):
        self.url = url
        if not msg:
            msg = "Invalid URL: {0}".format(url)
        super(InvalidURL, self).__init__(msg)


def _parse_url(url=None):
    """Split the path up into useful parts: bucket, obj_key"""
    if url is None:
        return ('', '')

    scheme, netloc, path, _, _ = parse.urlsplit(url)

    if scheme != 's3':
        raise InvalidURL(url, "URL scheme must be s3://")

    if path and not netloc:
        raise InvalidURL(url)

    return netloc, path[1:]


class S3(AwsApiClient):
    """Interface for managing S3 buckets. (API Version 2006-03-01)"""
    service_name = "s3"

    def get(self, url=None, delimiter="/"):
        """Path is an s3 url. Ommiting the path or providing "s3://" as the
        path will return a list of all buckets. Otherwise, all subdirectories
        and their contents will be shown.
        """
        params = {'Delimiter': delimiter}
        bucket, obj_key = _parse_url(url)

        if bucket:
            params['Bucket'] = bucket
        else:
            return self.call("ListBuckets", response_data_key="Buckets")

        if obj_key:
            params['Prefix'] = obj_key

        objects = self.call("ListObjects", response_data_key="Contents",
                            **params)
        if objects:
            for obj in objects:
                obj['url'] = "s3://{0}/{1}".format(bucket, obj['Key'])

        return objects

    def create(self, url):
        """Create a bucket, directory, or empty file."""
        bucket, obj_key = _parse_url(url)

        if not bucket:
            raise InvalidURL(url,
                             "You must specify a bucket and (optional) path")

        if obj_key:
            target = "/".join((bucket, obj_key))
        else:
            target = bucket

        return self.call("CreateBucket", bucket=target)

    def destroy(self, url, recursive=False):
        """Destroy a bucket, directory, or file. Specifying recursive=True
        recursively deletes all subdirectories and files."""
        bucket, obj_key = _parse_url(url)

        if not bucket:
            raise InvalidURL(url,
                             "You must specify a bucket and (optional) path")

        if obj_key:
            target = "/".join((bucket, obj_key))
        else:
            target = bucket

        if recursive:
            for obj in self.get(url, delimiter=''):
                self.destroy(obj['url'])

        return self.call("DeleteBucket", bucket=target)

    def upload(self, local_path, remote_url):
        """Copy a local file to an S3 location."""
        bucket, key = _parse_url(remote_url)

        with open(local_path, 'rb') as fp:
            return self.call("PutObject", bucket=bucket, key=key, body=fp)

    def download(self, remote_url, local_path, buffer_size=8 * 1024):
        """Copy S3 data to a local file."""
        bucket, key = _parse_url(remote_url)

        response_file = self.call("GetObject", bucket=bucket, key=key)['Body']
        with open(local_path, 'wb') as fp:
            buf = response_file.read(buffer_size)
            while buf:
                fp.write(buf)
                buf = response_file.read(buffer_size)

    def copy(self, src_url, dst_url):
        """Copy an S3 object to another S3 location."""
        src_bucket, src_key = _parse_url(src_url)
        dst_bucket, dst_key = _parse_url(dst_url)

        if not dst_bucket:
            dst_bucket = src_bucket
        params = {
            'copy_source': '/'.join((src_bucket, src_key)),
            'bucket': dst_bucket,
            'key': dst_key,
        }
        return self.call("CopyObject", **params)

    def move(self, src_url, dst_url):
        """Copy a single S3 object to another S3 location, then delete the
        original object."""
        self.copy(src_url, dst_url)
        self.destroy(src_url)



from acky.api import (
    AwsCollection,
    AwsApiClient,
)
from hashlib import md5


class QueueNotFound(Exception):
    pass


class SQSApiClient(AwsApiClient):
    service_name = "sqs"


class SQS(SQSApiClient):
    @property
    def Queues(self):
        return QueuesCollection(self._aws)

    @property
    def Messages(self):
        return MessagesCollection(self._aws)


class QueuesCollection(AwsCollection, SQSApiClient):
    def get(self, queue_name=None, prefix_match=False):
        # returns [QueueUrl, ...]
        # ListQueues, GetQueueUrl (no native filter)
        params = {}
        if queue_name:
            if prefix_match:
                params["queue_name_prefix"] = queue_name
            else:
                # if it's already a URL, just send it back as-is
                if "://" in queue_name:
                    return [queue_name]
                params["queue_name"] = queue_name
                return [self.call("GetQueueUrl", response_data_key="QueueUrl",
                        **params)]
        return self.call("ListQueues", response_data_key="QueueUrls", **params)

    def get_attributes(self, queue_url):
        return self.call("GetQueueAttributes",
                         response_data_key="Attributes",
                         queue_url=queue_url,
                         attribute_names=['All'])

    def create(self, name, attributes=None):
        # returns QueueUrl
        # CreateQueue
        params = {'queue_name': name}
        if attributes:
            params['attributes'] = attributes
        return self.call("CreateQueue", response_data_key="QueueUrl", **params)

    def destroy(self, queue):
        # returns bool
        # DeleteDBSnapshot
        queue_url = self.get(queue)[0]
        self.call("DeleteQueue", queue_url=queue_url)
        return True


class MessagesCollection(AwsCollection, SQSApiClient):
    def get(self, queue, max_messages=10, wait_time_seconds=0, filters=None,
            attributes=None):
        # returns [{<Message>}, ...]
        # ReceiveMessage (no native filter)
        if filters:
            raise Warning("Filters are ignored for this call")
        if '://' not in queue:
            queue_urls = self._aws.sqs.Queues.get(queue)
            if not queue_urls:
                raise QueueNotFound(queue)
            queue = queue_urls[0]
        if not attributes:
            attributes = ['All']
        params = {
            "queue_url": queue,
            "max_number_of_messages": max_messages,
            "wait_time_seconds": wait_time_seconds,
            "attribute_names": attributes,
        }
        return self.call("ReceiveMessage", response_data_key="Messages",
                         **params)

    def create(self, queue, message_body, verify=False):
        # returns message_id
        # SendMessage
        if '://' not in queue:
            queue_urls = self._aws.sqs.Queues.get(queue)
            if not queue_urls:
                raise QueueNotFound(queue)
            queue = queue_urls[0]
        params = {
            'queue_url': queue,
            'message_body': message_body,
        }
        result = self.call("SendMessage", **params)
        if verify:
            assert md5(message_body).hexdigest() == result['MD5OfMessageBody']
        return result['MessageId']

    def destroy(self, queue, receipt_handle):
        # returns bool
        # DeleteMessage
        if '://' not in queue:
            queue_urls = self._aws.sqs.Queues.get(queue)
            if not queue_urls:
                raise QueueNotFound(queue)
            queue = queue_urls[0]
        self.call("DeleteMessage", queue_url=queue,
                  receipt_handle=receipt_handle)
        return True



from acky.api import AwsApiClient


class STSApiClient(AwsApiClient):
    service_name = "sts"


class STS(STSApiClient):
    def GetFederationToken(self, name, policy='', duration_seconds=3600):
        params = {
            "name": name,
            "duration_seconds": duration_seconds,
        }
        if policy:
            params['policy'] = policy
        return self.call("GetFederationToken", **params)

    def GetSessionToken(self, duration_seconds=3600):
        params = {
            "duration_seconds": duration_seconds,
        }
        return self.call("GetSessionToken", **params)



from acky.api import (
    AwsCollection,
    AwsApiClient,
    make_filters,
)
from itertools import chain


class EC2ApiClient(AwsApiClient):
    service_name = "ec2"


class EC2(EC2ApiClient):

    def regions(self, continent='us', include_gov=False):
        # returns (string, ...)
        # DescribeRegions
        regions = self.call("DescribeRegions", response_data_key="Regions")
        if regions and continent and continent != "all":
            regions = [r for r in regions
                       if r['RegionName'].startswith("{}-".format(continent))]
        return regions

    def zones(self, region):
        # returns (string, ...)
        # DescribeAvailabilityZones
        raise NotImplementedError("aws.ec2.zones")

    @property
    def environment(self):
        env = super(EC2, self).environment
        env['hoster'] = 'ec2'
        return env

    @property
    def ACLs(self):
        return ACLCollection(self._aws)

    @property
    def ACEs(self):
        return ACECollection(self._aws)

    @property
    def ElasticIPs(self):
        return ElasticIPCollection(self._aws)

    @property
    def Instances(self):
        return InstanceCollection(self._aws)

    @property
    def SecurityGroups(self):
        return SecurityGroupCollection(self._aws)

    @property
    def IpPermissions(self):
        return IpPermissionsCollection(self._aws)

    @property
    def Volumes(self):
        return VolumeCollection(self._aws)

    @property
    def Snapshots(self):
        return SnapshotCollection(self._aws)

    @property
    def Subnets(self):
        return SubnetCollection(self._aws)

    @property
    def VPCs(self):
        return VPCCollection(self._aws)

    @property
    def PlacementGroups(self):
        return PlacementGroupCollection(self._aws)

    @property
    def KeyPairs(self):
        return KeyPairCollection(self._aws)

    @property
    def Tags(self):
        return TagCollection(self._aws)

    @property
    def Images(self):
        return ImageCollection(self._aws)


class ACLCollection(AwsCollection, EC2ApiClient):
    def get(self, filters=None):
        # returns (acl_info, ...)
        # DescribeNetworkAcls
        raise NotImplementedError()

    def create(self, vpc):
        # returns acl_info
        # CreateNetworkAcl
        raise NotImplementedError()

    def destroy(self, acl):
        # returns bool
        # DeleteNetworkAcl
        raise NotImplementedError()


class ACECollection(AwsCollection, EC2ApiClient):
    def get(self, filters=None):
        # returns (ace_info, ...)
        # DescribeNetworkAcls
        raise NotImplementedError()

    def add(self, acl, ace_list):
        # returns ace_info
        # CreateNetworkAclEntry
        raise NotImplementedError()

    def remove(self, acl, ace_list):
        # returns bool
        # DeleteNetworkAclEntry
        raise NotImplementedError()

    def replace(self, acl, old, new):
        # returns ace_info
        # CreateNetworkAclEntry, DeleteNetworkAclEntry
        raise NotImplementedError()


class ElasticIPCollection(AwsCollection, EC2ApiClient):
    """Interface to get, create, destroy, associate, and disassociate EIPs for
    classic EC2 domains and VPCs. (Amazon EC2 API Version 2014-06-15)
    """
    def get(self, filters=None):
        """List EIPs and associated information."""
        params = {}
        if filters:
            params["filters"] = make_filters(filters)
        return self.call("DescribeAddresses",
                         response_data_key="Addresses",
                         **params)

    def create(self, vpc=False):
        """Set vpc=True to allocate an EIP for a EC2-Classic instance.
        Set vpc=False to allocate an EIP for a VPC instance.
        """
        return self.call("AllocateAddress",
                         Domain="vpc" if vpc else "standard")

    def destroy(self, eip_or_aid, disassociate=False):
        """Release an EIP. If the EIP was allocated for a VPC instance, an
        AllocationId(aid) must be provided instead of a PublicIp. Setting
        disassociate to True will attempt to disassociate the IP before
        releasing it (required for associated nondefault VPC instances).
        """
        if "." in eip_or_aid:               # If an IP is given (Classic)
            # NOTE: EIPs are automatically disassociated for Classic instances.
            return "true" == self.call("ReleaseAddress",
                                       response_data_key="return",
                                       PublicIp=eip_or_aid)
        else:                               # If an AID is given (VPC)
            if disassociate:
                self.disassociate(eip_or_aid)
            return "true" == self.call("ReleaseAddress",
                                       response_data_key="return",
                                       AllocationId=eip_or_aid)

    def associate(self, eip_or_aid,
                  instance_id='', network_interface_id='', private_ip=''):
        """Associate an EIP with a given instance or network interface. If
        the EIP was allocated for a VPC instance, an AllocationId(aid) must
        be provided instead of a PublicIp.
        """
        if "." in eip_or_aid:               # If an IP is given (Classic)
            return self.call("AssociateAddress",
                             PublicIp=eip_or_aid,
                             InstanceId=instance_id,
                             NetworkInterfaceId=network_interface_id,
                             PrivateIpAddress=private_ip)
        else:                               # If an AID is given (VPC)
            return self.call("AssociateAddress",
                             AllocationId=eip_or_aid,
                             InstanceId=instance_id,
                             NetworkInterfaceId=network_interface_id,
                             PrivateIpAddress=private_ip)

    def disassociate(self, eip_or_aid):
        """Disassociates an EIP. If the EIP was allocated for a VPC instance,
        an AllocationId(aid) must be provided instead of a PublicIp.
        """
        if "." in eip_or_aid:               # If an IP is given (Classic)
            return "true" == self.call("DisassociateAddress",
                                       response_data_key="return",
                                       PublicIp=eip_or_aid)
        else:                               # If an AID is given (VPC)
            return "true" == self.call("DisassociateAddress",
                                       response_data_key="return",
                                       AllocationId=eip_or_aid)


class InstanceCollection(AwsCollection, EC2ApiClient):

    def get(self, instance_ids=None, filters=None):
        """List instance info."""
        params = {}
        if filters:
            params["filters"] = make_filters(filters)
        if instance_ids:
            params['InstanceIds'] = instance_ids
        reservations = self.call("DescribeInstances",
                                 response_data_key="Reservations",
                                 **params)
        if reservations:
            return list(chain(*(r["Instances"] for r in reservations)))
        return []

    def create(self, ami, count, config=None):
        """Create an instance using the launcher."""
        return self.Launcher(config=config).launch(ami, count)

    def destroy(self, instance_id):
        """Terminate a single given instance."""
        return self.control(instance_id, "terminate")

    def control(self, instances, action):
        """Valid actions: start, stop, reboot, terminate, protect, and
        unprotect.
        """
        if not isinstance(instances, list) and\
           not isinstance(instances, tuple):
            instances = [instances]
        actions = {'start': {'operation': "StartInstances",
                             'response_data_key': "StartingInstances",
                             'InstanceIds': instances},
                   'stop': {'operation': "StopInstances",
                            'response_data_key': "StoppingInstances",
                            'InstanceIds': instances},
                   'reboot': {'operation': "RebootInstances",
                              'response_data_key': "return",
                              'InstanceIds': instances},
                   'terminate': {'operation': "TerminateInstances",
                                 'response_data_key': "TerminatingInstances",
                                 'InstanceIds': instances},
                   'protect': {'operation': "ModifyInstanceAttribute",
                               'response_data_key': "return",
                               'Attribute': 'disableApiTermination',
                               'Value': 'true'},
                   'unprotect': {'operation': "ModifyInstanceAttribute",
                                 'response_data_key': "return",
                                 'Attribute': 'disableApiTermination',
                                 'Value': 'false'}}
        if (action in ('protect', 'unprotect')):
            for instance in instances:
                self.call(InstanceId=instance, **actions[action])
            return "true"
        else:
            return self.call(**actions[action])

    def Launcher(self, config=None):
        """Provides a configurable launcher for EC2 instances."""
        class _launcher(EC2ApiClient):
            """Configurable launcher for EC2 instances. Create the Launcher
            (passing an optional dict of its attributes), set its attributes
            (as described in the RunInstances API docs), then launch().
            """
            def __init__(self, aws, config):
                super(_launcher, self).__init__(aws)
                self.config = config
                self._attr = list(self.__dict__.keys()) + ['_attr']

            def launch(self, ami, min_count, max_count=0):
                """Use given AMI to launch min_count instances with the
                current configuration. Returns instance info list.
                """
                params = config.copy()
                params.update(dict([i for i in self.__dict__.items()
                                    if i[0] not in self._attr]))
                return self.call("RunInstances",
                                 ImageId=ami,
                                 MinCount=min_count,
                                 MaxCount=max_count or min_count,
                                 response_data_key="Instances",
                                 **params)

        if not config:
            config = {}
        return _launcher(self._aws, config)

    def status(self, all_instances=None, instance_ids=None, filters=None):
        """List instance info."""
        params = {}
        if filters:
            params["filters"] = make_filters(filters)
        if instance_ids:
            params['InstanceIds'] = instance_ids
        if all_instances is not None:
            params['IncludeAllInstances'] = all_instances
        statuses = self.call("DescribeInstanceStatus",
                             response_data_key="InstanceStatuses",
                             **params)
        return statuses

    def events(self, all_instances=None, instance_ids=None, filters=None):
        """a list of tuples containing instance Id's and event information"""
        params = {}
        if filters:
            params["filters"] = make_filters(filters)
        if instance_ids:
            params['InstanceIds'] = instance_ids
        statuses = self.status(all_instances, **params)
        event_list = []
        for status in statuses:
            if status.get("Events"):
                for event in status.get("Events"):
                    event[u"InstanceId"] = status.get('InstanceId')
                    event_list.append(event)
        return event_list


class KeyPairCollection(AwsCollection, EC2ApiClient):
    def get(self, filters=None):
        """List key info."""
        params = {}
        if filters:
            params["filters"] = make_filters(filters)
        return self.call("DescribeKeyPairs",
                         response_data_key="KeyPairs",
                         **params)

    def create(self, key_name):
        """Create a new key with a given name."""
        return self.call("CreateKeyPair", KeyName=key_name)

    def destroy(self, key_name):
        """Delete a key."""
        return self.call("DeleteKeyPair", KeyName=key_name)


class PlacementGroupCollection(AwsCollection, EC2ApiClient):
    def get(self, filters=None):
        # returns (sg_info, ...)
        # DescribePlacementGroups
        params = {}
        if filters:
            params["filters"] = make_filters(filters)
        return self.call("DescribePlacementGroups",
                         response_data_key="PlacementGroups",
                         **params)

    def create(self, group_name, strategy="cluster"):
        # returns sg_info
        params = {
            "strategy": strategy
        }
        # CreatePlacementGroup
        if callable(group_name):
            params['group_name'] = group_name(self.environment)
        else:
            params['group_name'] = group_name
        return self.call("CreatePlacementGroup", **params)

    def destroy(self, pg):
        # returns bool
        # DeletePlacementGroup
        return self.call("DeletePlacementGroup", group_name=pg)


class SecurityGroupCollection(AwsCollection, EC2ApiClient):
    def get(self, filters=None, exclude_vpc=False):
        # returns (sg_info, ...)
        # DescribeSecurityGroups
        params = {}
        if filters:
            params["filters"] = make_filters(filters)
        groups = self.call("DescribeSecurityGroups",
                           response_data_key="SecurityGroups",
                           **params)
        if groups and exclude_vpc:
            # Exclude any group that belongs to a VPC
            return [g for g in groups if not g.get('VpcId')]
        else:
            return groups

    def create(self, name, description, vpc=None):
        # returns sg_info
        params = {
            "Description": description,
        }
        # CreateSecurityGroup
        if callable(name):
            params['GroupName'] = name(self.environment)
        else:
            params['GroupName'] = name
        if vpc:
            params["VpcId"] = vpc
        return self.call("CreateSecurityGroup", **params)

    def destroy(self, sg):
        # returns bool
        # DeleteSecurityGroup
        return self.call("DeleteSecurityGroup", GroupId=sg)


class IpPermissionsCollection(AwsCollection, EC2ApiClient):
    def get(self, filters=None):
        # returns (sgr_info, ...)
        # DescribeSecurityGroups
        raise NotImplementedError()

    def modify(self, api_action, sgid, other, proto_spec):
        """Make a change to a security group. api_action is an EC2 API name.
           Other is one of:
            - a group (sg-nnnnnnnn)
            - a group with account (<user id>/sg-nnnnnnnn)
            - a CIDR block (n.n.n.n/n)
           Proto spec is a triplet (<proto>, low_port, high_port)."""
        params = {'group_id': sgid, 'ip_permissions': []}
        perm = {}
        params['ip_permissions'].append(perm)

        proto, from_port, to_port = proto_spec
        perm['IpProtocol'] = proto
        perm['FromPort'] = from_port or 0
        perm['ToPort'] = to_port or from_port or 65535

        if other.startswith("sg-"):
            perm['UserIdGroupPairs'] = [{'GroupId': other}]
        elif "/sg-" in other:
            account, group_id = other.split("/", 1)
            perm['UserIdGroupPairs'] = [{
                'UserId': account,
                'GroupId': group_id,
            }]
        else:
            perm['IpRanges'] = [{'CidrIp': other}]

        return self.call(api_action, **params)

    def add(self, sgid, other, proto_spec, direction="in"):
        """Add a security group rule to group <sgid>.
           Direction is either 'in' (ingress) or 'out' (egress).
           See modify() for other parameters."""
        # returns bool
        # AuthorizeSecurityGroupIngress, AuthorizeSecurityGroupEgress
        if direction == "in":
            api = "AuthorizeSecurityGroupIngress"
        elif direction == "out":
            api = "AuthorizeSecurityGroupEgress"
        else:
            raise ValueError("direction must be one of ('in', 'out')")
        return self.modify(api, sgid, other, proto_spec)

    def remove(self, sgid, other, proto_spec, direction="in"):
        """Remove a security group rule from group <sgid>.
           Direction is either 'in' (ingress) or 'out' (egress).
           See modify() for other parameters."""
        # returns (removed_sgr_info, ...)
        # RevokeSecurityGroupIngress, RevokeSecurityGroupEgress
        if direction == "in":
            api = "RevokeSecurityGroupIngress"
        elif direction == "out":
            api = "RevokeSecurityGroupEgress"
        else:
            raise ValueError("direction must be one of ('in', 'out')")
        return self.modify(api, sgid, other, proto_spec)


class VolumeCollection(AwsCollection, EC2ApiClient):
    """Interface to get, create, destroy, and attach for EBS Volumes.
    (Amazon EC2 API Version 2014-06-15)
    """
    def get(self, volume_ids=None, filters=None):
        """List EBS Volume info."""
        params = {}
        if filters:
            params["filters"] = make_filters(filters)
        if isinstance(volume_ids, str):
            volume_ids = [volume_ids]
        return self.call("DescribeVolumes",
                         VolumeIds=volume_ids,
                         response_data_key="Volumes",
                         **params)

    def create(self, az, size_or_snap, volume_type=None, iops=None,
               encrypted=True):
        """Create an EBS Volume using an availability-zone and size_or_snap
        parameter, encrypted by default.
        If the volume is crated from a snapshot, (str)size_or_snap denotes
        the snapshot id. Otherwise, (int)size_or_snap denotes the amount of
        GiB's to allocate. iops must be set if the volume type is io1.
        """
        kwargs = {}
        kwargs['encrypted'] = encrypted
        if volume_type:
            kwargs['VolumeType'] = volume_type
        if iops:
            kwargs['Iops'] = iops
        is_snapshot_id = False
        try:
            size_or_snap = int(size_or_snap)
        except ValueError:
            is_snapshot_id = True
        if is_snapshot_id:
            return self.call("CreateVolume", AvailabilityZone=az,
                             SnapshotId=size_or_snap, **kwargs)
        return self.call("CreateVolume", AvailabilityZone=az,
                         Size=size_or_snap, **kwargs)

    def destroy(self, volume_id):
        """Delete a volume by volume-id and return success boolean."""
        return 'true' == self.call("DeleteVolume", VolumeId=volume_id,
                                   response_data_key="return")

    def attach(self, volume_id, instance_id, device_path):
        """Attach a volume to an instance, exposing it with a device name."""
        return self.call("AttachVolume",
                         VolumeId=volume_id, InstanceId=instance_id,
                         Device=device_path)

    def detach(self, volume_id, instance_id='', device_path='', force=False):
        """Detach a volume from an instance."""
        return self.call("DetachVolume",
                         VolumeId=volume_id, InstanceId=instance_id,
                         Device=device_path, force=force)


class SnapshotCollection(AwsCollection, EC2ApiClient):
    def get(self, filters=None):
        # returns (snap_info, ...)
        # DescribeSnapshots
        params = {}
        if filters:
            params["filters"] = make_filters(filters)
        return self.call("DescribeSnapshots",
                         response_data_key="Snapshots",
                         **params)

    def create(self, volume_id, description=None):
        # returns snap_info
        # CreateSnapshot
        return self.call("CreateSnapshot",
                         VolumeId=volume_id,
                         Description=description)

    def destroy(self, snapshot_id):
        # returns bool
        # DeleteSnapshot
        return self.call("DeleteSnapshot", SnapshotId=snapshot_id)


class SubnetCollection(AwsCollection, EC2ApiClient):
    def get(self, filters=None):
        # returns (subnet_info, ...)
        # DescribeSubnets
        params = {}
        if filters:
            params["filters"] = make_filters(filters)
        return self.call("DescribeSubnets",
                         response_data_key="Subnets",
                         **params)

    def create(self, vpc_id, cidr, availability_zone):
        # returns subnet_info
        # CreateSubnet
        return self.call("CreateSubnet",
                         VpcId=vpc_id,
                         CidrBlock=cidr,
                         response_data_key="Subnet")

    def destroy(self, subnet_id):
        # returns bool
        # DeleteSubnet
        if self.call("DeleteSubnet", SubnetId=subnet_id,
                     response_data_key="return"):
            return True
        return False


class VPCCollection(AwsCollection, EC2ApiClient):
    def get(self, filters=None):
        # returns (vpc_info, ...)
        # DescribeVpcs
        params = {}
        if filters:
            params["filters"] = make_filters(filters)
        return self.call("DescribeVpcs", response_data_key="Vpcs", **params)

    def create(self, cidr, tenancy="default"):
        # returns vpc_info
        # CreateVpc
        raise NotImplementedError()

    def destroy(self, vpc):
        # returns bool
        # DeleteVpc
        raise NotImplementedError()


class TagCollection(AwsCollection, EC2ApiClient):
    def get(self, filters=None):
        # returns (tag_info, ...)
        # DescribeTags
        params = {}
        if filters:
            params["filters"] = make_filters(filters)
        return self.call("DescribeTags",
                         response_data_key="Tags",
                         **params)

    def create(self, resource_ids, tags):
        # returns bool
        # CreateTags
        return self.call("CreateTags", resources=resource_ids, tags=tags)

    def destroy(self, resource_ids, tags):
        # returns bool
        # DeleteTags
        return self.call("DeleteTags", resources=resource_ids, tags=tags)


class ImageCollection(AwsCollection, EC2ApiClient):
    def get(self, image_ids=None, owners=None, executable_users=None, filters=None):
        # returns (image_info, ...)
        # DescribeImages
        params = {}
        if filters:
            params["filters"] = make_filters(filters)
        if image_ids:
            params["ImageIds"] = image_ids
        if owners:
            params["Owners"] = owners
        if executable_users:
            params["ExecutableUsers"] = executable_users
        return self.call("DescribeImages",
                         response_data_key="Images",
                         **params)

    def create(self, instance_id, name, no_reboot=True, description=None, block_device_mappings=None):
        # returns image_id
        # CreateImage
        params = {
            "InstanceId": instance_id,
            "Name": name,
            "NoReboot": no_reboot
        }
        if description:
            params["Description"] = description
        if block_device_mappings:
            params["BlockDeviceMappings"] = block_device_mappings
        return self.call("CreateImage",
                         response_data_key="ImageId",
                         **params)

    def destroy(self, image_id):
        # returns bool
        # CreateImage
        return self.call("DeregisterImage", ImageId=image_id)



"""Classes to represent an Amazon Web Services connection and its objects"""
from datetime import datetime
import acky.ec2
import acky.iam
import acky.rds
import acky.sts
import acky.sqs
import acky.s3
import botocore.session


class AWS(object):
    def __init__(self, region, profile=None):
        env_vars = {
            'region': ('region', 'BOTO_DEFAULT_REGION', region),
            'profile': (None, 'BOTO_DEFAULT_PROFILE', profile),
        }
        self.session = botocore.session.get_session(env_vars)
        self.session.profile = profile
        self.region = region

    @property
    def userinfo(self):
        if not hasattr(self, '_user'):
            self._user = self.iam.Users.get_current()
        return self._user

    @property
    def username(self):
        return self.userinfo['UserName']

    @property
    def account_id(self):
        arn = self.userinfo['Arn']
        return arn.split(":")[4]

    @property
    def environment(self):
        return {
            'region': self.region,
            'username': self.username,
            'account': self.account_id,
            'date': datetime.isoformat(datetime.now()),
        }

    @property
    def ec2(self):
        return acky.ec2.EC2(self)

    @property
    def iam(self):
        return acky.iam.IAM(self)

    @property
    def rds(self):
        return acky.rds.RDS(self)

    @property
    def sqs(self):
        return acky.sqs.SQS(self)

    @property
    def sts(self):
        return acky.sts.STS(self)

    @property
    def s3(self):
        return acky.s3.S3(self)



import botocore
import unittest
import acky
try:
    from unittest.mock import patch, call, ANY
except ImportError:
    from mock import patch, call, ANY

GENERIC_AID = "aid-123"
GENERIC_AMI = "ami-123"
GENERIC_SUBNET = "subnet-test"
GENERIC_IP = "8.8.8.8"
GENERIC_INSTANCE = "i-123"
GENERIC_VOL = "vol-123"
GENERIC_AMI_DESCRIPTION = "{0} description".format(GENERIC_AMI)

class _AWS(object):
    """AWS object for mock testing with only basic features."""
    def __init__(self):
        self.session = botocore.session.get_session()
        self.region = 'us-east-1'

    @property
    def ec2(self):
        return acky.ec2.EC2(self)


class TestRegions(unittest.TestCase):
    @patch('acky.api.AwsApiClient.call')
    def test_regions(self, _call):
        _AWS().ec2.regions()
        _call.assert_called_once_with("DescribeRegions",
                                      response_data_key=ANY)


class _TestEC2Collection():
    class_name = None
    commands = None
    get_args = [{}]
    get_expectations = None
    create_args = ['test']
    create_expectations = None
    destroy_args = ['test']
    destroy_expectations = None
    instance = None

    @patch('acky.api.AwsApiClient.call')
    def test_get(self, _call):
        """Generic testing method for get() calls. Modify get_args and
        get_expectations for better coverage.
        """
        for args, expectation in zip(self.get_args,
                                     self.get_expectations):
            try:
                self.instance.get(**args)
            except NotImplementedError:
                continue
            assert expectation in _call.mock_calls,\
                "Incorrect get call by {} (expected {} in {})".\
                format(self.class_name, expectation,
                       _call.mock_calls)

    @patch('acky.api.AwsApiClient.call')
    def test_create(self, _call):
        """Generic testing method for create() calls. Modify create_args and
        create_expectations for better coverage.
        """
        for args, expectation in zip(self.create_args,
                                     self.create_expectations):
            try:
                self.instance.create(**args)
            except NotImplementedError:
                continue
            assert expectation in _call.mock_calls,\
                "Incorrect create call by {} (expected {} in {})".\
                format(self.class_name, expectation,
                       _call.mock_calls)

    @patch('acky.api.AwsApiClient.call')
    def test_destroy(self, _call):
        """Generic testing method for destroy() calls. Modify destroy_args
        and destroy_expectations for better coverage.
        """
        for args, expectation in zip(self.destroy_args,
                                     self.destroy_expectations):
            try:
                self.instance.destroy(**args)
            except NotImplementedError:
                continue
            assert expectation in _call.mock_calls,\
                "Incorrect destroy call by {} (expected {} in {})".\
                format(self.class_name, expectation,
                       _call.mock_calls)


class TestACLCollection(_TestEC2Collection, unittest.TestCase):
    """Not implemented yet."""

    class_name = "ACLs"
    commands = {'get': 'DescribeNetworkAcls',
                'create': 'CreateNetworkAcl',
                'destroy': 'DeleteNetworkAcl'}
    get_expectations = [{}]
    create_args = [{'vpc': "test"}]
    create_expectations = [{}]
    destroy_args = [{'acl': "test"}]
    destroy_expectations = [{}]
    instance = _AWS().ec2.ACLs

'''
class TestACECollection(_TestEC2Collection, unittest.TestCase):
    """Not implemented yet."""
    class_name = "ACE"
    commands = {'get': 'DescribeNetworkAcls',
                'create': 'CreateNetworkAclEntry',
                'destroy': 'DeleteNetworkAclEntry'}
    get_expectations = [{}]
    create_args = [{}]
    create_expectations = [{}]
    destroy_args = [{}]
    destroy_expectations = [{}]
    instance = _AWS().ec2.ACEs
'''


class ElasticIPCollection(_TestEC2Collection, unittest.TestCase):
    class_name = "EIPs"
    commands = {'get': 'DescribeAddresses',
                'create': 'AllocateAddress',
                'destroy': 'ReleaseAddress'}
    get_expectations = [call(commands['get'],
                             response_data_key="Addresses")]
    create_args = [{'vpc': True}, {'vpc': False}]
    create_expectations = [call(commands['create'],
                                Domain="vpc"),
                           call(commands['create'],
                                Domain="standard")]
    destroy_args = [{'eip_or_aid': GENERIC_IP},
                    {'eip_or_aid': GENERIC_AID, 'disassociate': False},
                    {'eip_or_aid': GENERIC_AID, 'disassociate': True}]
    destroy_expectations = [call(commands['destroy'],
                                 PublicIp='8.8.8.8',
                                 response_data_key="return"),
                            call(commands['destroy'],
                                 response_data_key="return",
                                 AllocationId=GENERIC_AID),
                            call(commands['destroy'],
                                 response_data_key="return",
                                 AllocationId=GENERIC_AID)]
    instance = _AWS().ec2.ElasticIPs

    @patch('acky.api.AwsApiClient.call')
    def test_associate(self, _call):
        associate_command = "AssociateAddress"
        associate_args = [{'eip_or_aid': GENERIC_IP, 'instance_id': "i-test"},
                          {'eip_or_aid': GENERIC_AID, 'instance_id': "i-test"},
                          {'eip_or_aid': GENERIC_AID,
                           'network_interface_id': "net-test"}]
        associate_expectations = [call(associate_command,
                                       PublicIp=GENERIC_IP,
                                       InstanceId="i-test",
                                       NetworkInterfaceId='',
                                       PrivateIpAddress=''),
                                  call(associate_command,
                                       AllocationId=GENERIC_AID,
                                       InstanceId="i-test",
                                       NetworkInterfaceId='',
                                       PrivateIpAddress=''),
                                  call(associate_command,
                                       AllocationId=GENERIC_AID,
                                       InstanceId='',
                                       NetworkInterfaceId="net-test",
                                       PrivateIpAddress='')]
        for args, expectation in zip(associate_args,
                                     associate_expectations):
            self.instance.associate(**args)
            assert expectation in _call.mock_calls,\
                "Incorrect associate call by {} (expected {} in {})".\
                format(self.class_name, expectation,
                       _call.mock_calls)

    @patch('acky.api.AwsApiClient.call')
    def test_disassociate(self, _call):
        disassociate_command = "DisassociateAddress"
        disassociate_args = [{'eip_or_aid': GENERIC_IP},
                             {'eip_or_aid': GENERIC_AID}]
        disassociate_expectations = [call(disassociate_command,
                                          response_data_key="return",
                                          PublicIp=GENERIC_IP),
                                     call(disassociate_command,
                                          response_data_key="return",
                                          AllocationId=GENERIC_AID)]
        for args, expectation in zip(disassociate_args,
                                     disassociate_expectations):
            self.instance.disassociate(**args)
            assert expectation in _call.mock_calls,\
                "Incorrect disassociate call by {} (expected {} in {})".\
                format(self.class_name, expectation,
                       _call.mock_calls)


class TestInstanceCollection(_TestEC2Collection, unittest.TestCase):
    class_name = "Instances"
    commands = {'get': 'DescribeInstances',
                'create': 'RunInstances',
                'destroy': 'TerminateInstances'}
    get_expectations = [call(commands['get'],
                             response_data_key="Reservations")]
    create_args = [{'ami': GENERIC_AMI, 'count': 1}]
    create_expectations = [call('RunInstances', MinCount=1,
                                ImageId=GENERIC_AMI,
                                response_data_key='Instances', MaxCount=1)]
    destroy_args = [{'instance_id': GENERIC_INSTANCE}]
    destroy_expectations = [call(operation='TerminateInstances',
                                 response_data_key='TerminatingInstances',
                                 InstanceIds=[GENERIC_INSTANCE])]
    instance = _AWS().ec2.Instances

    @patch('acky.api.AwsApiClient.call')
    def test_control(self, _call):
        ctrl_data = {'start': {'operation': "StartInstances",
                               'response_data_key': "StartingInstances",
                               'InstanceIds': [GENERIC_INSTANCE]},
                     'stop': {'operation': "StopInstances",
                              'response_data_key': "StoppingInstances",
                              'InstanceIds': [GENERIC_INSTANCE]},
                     'reboot': {'operation': "RebootInstances",
                                'response_data_key': "return",
                                'InstanceIds': [GENERIC_INSTANCE]},
                     'terminate': {'operation': "TerminateInstances",
                                   'response_data_key': "TerminatingInstances",
                                   'InstanceIds': [GENERIC_INSTANCE]},
                     'protect': {'operation': "ModifyInstanceAttribute",
                                 'response_data_key': "return",
                                 'Attribute': 'disableApiTermination',
                                 'Value': 'true'},
                     'unprotect': {'operation': "ModifyInstanceAttribute",
                                   'response_data_key': "return",
                                   'Attribute': 'disableApiTermination',
                                   'Value': 'false'}}
        control_args = [{'instances': GENERIC_INSTANCE, 'action': "start"},
                        {'instances': GENERIC_INSTANCE, 'action': "stop"},
                        {'instances': GENERIC_INSTANCE, 'action': "protect"},
                        {'instances': GENERIC_INSTANCE, 'action': "unprotect"}]
        control_expectations = [call(operation=ctrl_data['start']['operation'],
                                     response_data_key=
                                     ctrl_data['start']['response_data_key'],
                                     InstanceIds=[GENERIC_INSTANCE]),
                                call(operation=ctrl_data['stop']['operation'],
                                     response_data_key='StoppingInstances',
                                     InstanceIds=['i-123']),
                                call(InstanceId='i-123',
                                     Attribute='disableApiTermination',
                                     operation=
                                     ctrl_data['protect']['operation'],
                                     response_data_key='return', Value='true'),
                                call(InstanceId='i-123',
                                     Attribute='disableApiTermination',
                                     operation=
                                     ctrl_data['protect']['operation'],
                                     response_data_key='return',
                                     Value='false')]
        for args, expectation in zip(control_args,
                                     control_expectations):
            self.instance.control(**args)
            assert expectation in _call.mock_calls,\
                "Incorrect control call by {} (expected {} in {})".\
                format(self.class_name, expectation,
                       _call.mock_calls)


class TestKeyCollection(_TestEC2Collection, unittest.TestCase):
    class_name = "KeyPairs"
    commands = {'get': 'DescribeKeyPairs',
                'create': 'CreateKeyPair',
                'destroy': 'DeleteKeyPair'}
    get_expectations = [call('DescribeKeyPairs', response_data_key='KeyPairs')]
    create_args = [{'key_name': "test-key"}]
    create_expectations = [call('CreateKeyPair', KeyName='test-key')]
    destroy_args = [{'key_name': "test-key"}]
    destroy_expectations = [call('DeleteKeyPair', KeyName='test-key')]
    instance = _AWS().ec2.KeyPairs


class TestPlacementGroupCollection(_TestEC2Collection, unittest.TestCase):
    class_name = "PlacementGroups"
    commands = {'get': 'DescribePlacementGroups',
                'create': 'CreatePlacementGroup',
                'destroy': 'DeletePlacementGroup'}
    get_expectations = [call('DescribePlacementGroups',
                             response_data_key='PlacementGroups')]
    create_args = [{'group_name': 'test-group'}]
    create_expectations = [call('CreatePlacementGroup',
                                group_name='test-group', strategy='cluster')]
    destroy_args = [{'pg': 'test-group'}]
    destroy_expectations = [call('DeletePlacementGroup',
                                 group_name='test-group')]
    instance = _AWS().ec2.PlacementGroups


class TestSecurityGroupCollection(_TestEC2Collection, unittest.TestCase):
    class_name = "SecurityGroups"
    commands = {'get': 'DescribeSecurityGroups',
                'create': 'CreateSecurityGroup',
                'destroy': 'DeleteSecurityGroup'}
    get_expectations = [call(commands['get'],
                             response_data_key='SecurityGroups')]
    create_args = [{'name': "test-group", 'description': "test-description"}]
    create_expectations = [call(commands['create'], GroupName='test-group',
                                Description='test-description')]
    destroy_args = [{'sg': "test-group"}]
    destroy_expectations = [call(commands['destroy'], GroupId='test-group')]
    instance = _AWS().ec2.SecurityGroups


'''
class TestIpPermissionsCollection(_TestEC2Collection, unittest.TestCase):
    """Not full implemented yet."""
    class_name = "IpPermissions"
    commands = {'get': '',
                'create': '',
                'destroy': ''}
    get_expectations = [{}]
    create_args = [{}]
    create_expectations = [{}]
    destroy_args = [{}]
    destroy_expectations = [{}]
    instance = _AWS().ec2.IpPermissions
'''


class TestVolumeCollection(_TestEC2Collection, unittest.TestCase):
    class_name = "Volumes"
    commands = {'get': 'DescribeVolumes',
                'create': 'CreateVolume',
                'destroy': 'DeleteVolume'}
    get_expectations = [
        call(commands['get'], VolumeIds=None, response_data_key='Volumes')
    ]
    create_args = [{'az': "us-east-1", 'size_or_snap': 8}]
    create_expectations = [call(commands['create'], encrypted=True,
                                AvailabilityZone='us-east-1', Size=8)]
    destroy_args = [{'volume_id': GENERIC_VOL}]
    destroy_expectations = [call(commands['destroy'],
                                 response_data_key='return',
                                 VolumeId='vol-123')]
    instance = _AWS().ec2.Volumes


class TestSnapshotCollection(_TestEC2Collection, unittest.TestCase):
    class_name = "Snapshots"
    commands = {'get': "DescribeSnapshots",
                'create': "CreateSnapshot",
                'destroy': "DeleteSnapshot"}
    get_expectations = [call(commands['get'],
                             response_data_key='Snapshots')]
    create_args = [{'volume_id': GENERIC_VOL}]
    create_expectations = [call(commands['create'],
                                Description=None, VolumeId='vol-123')]
    destroy_args = [{'snapshot_id': "snap-123"}]
    destroy_expectations = [call(commands['destroy'], SnapshotId='snap-123')]
    instance = _AWS().ec2.Snapshots


class TestSubnetCollection(_TestEC2Collection, unittest.TestCase):
    class_name = "Subnets"
    commands = {'get': "DescribeSubnets",
                'create': "CreateSubnet",
                'destroy': "DeleteSubnet"}
    get_expectations = [call('DescribeSubnets', response_data_key='Subnets')]
    create_args = [{'vpc_id': "vpc-test", 'cidr': "cidr-test",
                    'availability_zone': "us-east-4b"}]
    create_expectations = [call(commands['create'], VpcId='vpc-test',
                                CidrBlock='cidr-test',
                                response_data_key='Subnet')]
    destroy_args = [{'subnet_id': GENERIC_SUBNET}]
    destroy_expectations = [call(commands['destroy'], SubnetId='subnet-test',
                                 response_data_key='return')]
    instance = _AWS().ec2.Subnets


class TestVPCCollection(_TestEC2Collection, unittest.TestCase):
    """Not implemented."""

    class_name = "VPCs"
    commands = {'get': "DescribeVpcs",
                'create': "",
                'destroy': ""}
    get_expectations = [call(commands['get'], response_data_key='Vpcs')]
    create_args = [{'cidr': ""}]
    create_expectations = [{}]
    destroy_args = [{'vpc': ""}]
    destroy_expectations = [{}]
    instance = _AWS().ec2.VPCs


class TestTagCollection(_TestEC2Collection, unittest.TestCase):
    class_name = "Tags"
    commands = {'get': 'DescribeTags',
                'create': 'CreateTags',
                'destroy': 'DeleteTags'}
    get_expectations = [call(commands['get'], response_data_key='Tags')]
    create_args = [{'resource_ids': ["res-test"], 'tags': ["tag-test"]}]
    create_expectations = [call(commands['create'], resources=["res-test"],
                                tags=["tag-test"])]
    destroy_args = [{'resource_ids': ["res-test"], 'tags': ["tag-test"]}]
    destroy_expectations = [call(commands['destroy'], resources=["res-test"],
                                 tags=["tag-test"])]
    instance = _AWS().ec2.Tags


class TestImageCollection(_TestEC2Collection, unittest.TestCase):
    class_name = "Images"
    commands = {'get': 'DescribeImages',
                'create': 'CreateImage',
                'destroy': 'DeregisterImage'}
    get_expectations = [call(commands['get'], response_data_key='Images')]
    create_args = [{
        'instance_id': GENERIC_INSTANCE,
        'name': GENERIC_INSTANCE,
        'no_reboot':  True,
        'description': GENERIC_AMI_DESCRIPTION
    }]
    create_expectations = [call(
        commands['create'],
        response_data_key='ImageId',
        Name=GENERIC_INSTANCE,
        Description=GENERIC_AMI_DESCRIPTION,
        InstanceId=GENERIC_INSTANCE,
        NoReboot=True
    )]
    destroy_args = [{'image_id': GENERIC_AMI}]
    destroy_expectations = [call(commands['destroy'],
                                 ImageId=GENERIC_AMI)]
    instance = _AWS().ec2.Images



from __future__ import print_function
from setuptools import setup, find_packages
from acky import __version__
import sys

if sys.version_info <= (2, 7):
    error = "ERROR: acky requires Python 2.7 or later"
    print(error, file=sys.stderr)
    sys.exit(1)

with open('README.rst') as f:
    long_description = f.read()

setup(
    name="acky",
    version=__version__,
    description="A consistent API to AWS",
    long_description=long_description,
    author="Matthew Wedgwood",
    author_email="mw@rmn.com",
    url="http://github.com/RetailMeNot/acky",
    install_requires=[
        "botocore == 0.45.0",
    ],
    packages=find_packages(),
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Internet",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Topic :: Internet",
    ],
    license="MIT",
)



import logging
from xml.dom.minidom import parseString


def make_filters(data, key_name='Name', values_name='Values'):
    data = data.copy()
    for key, value in data.items():
        if isinstance(value, str) or not hasattr(value, '__iter__'):
            data[key] = [value]
    return make_map(data, key_name, values_name)


def make_map(data, key_name='Name', value_name='Value'):
    value_map = []
    for key, value in data.items():
        if not isinstance(value, str) and not hasattr(value, '__iter__'):
            value = str(value)
        value_map.append({key_name: key, value_name: value})
    return value_map


class AWSErrorNotFound(Exception):
    pass


def extract_aws_error(xml_string):
    def _bomb(xml_string):
        e = AWSErrorNotFound("The string provided does not appear to be an "
                             "XML AWS error response")
        e.xml_string = xml_string
        raise e

    try:
        dom = parseString(xml_string)
    except TypeError:
        _bomb(xml_string)

    for tag in dom.getElementsByTagName('Error'):
        children = [n.tagName for n in tag.childNodes
                    if n.nodeType != n.TEXT_NODE]
        if 'Code' in children and 'Message' in children:
            code = tag.getElementsByTagName('Code')[0].firstChild.data
            message = tag.getElementsByTagName('Message')[0].firstChild.data
            return code, message

    _bomb(xml_string)


class AWSServiceUnavailable(Exception):
    pass


class AWSCallError(Exception):
    def __init__(self, response, operation):
        self.response = response
        self.operation = operation
        self.args = (response, operation)
        self.code, self.message = extract_aws_error(response.text)

    def __str__(self):
        return "{}: {}: {}".format(self.operation,
                                   self.code,
                                   self.message)


class AwsApiClient(object):
    service_name = None

    def __init__(self, aws):
        self._aws = aws
        self._service = aws.session.get_service(self.service_name)
        self._endpoint = self._service.get_endpoint(aws.region)

    def call(self, operation, response_data_key=None, *args, **kwargs):
        log = logging.getLogger(__name__)
        op = self._service.get_operation(operation)
        log.debug("Calling {} action '{}'".format(self._service, operation))
        resp, data = op.call(self._endpoint, *args, **kwargs)
        if not resp.ok:
            raise AWSCallError(resp, operation)
        if response_data_key:
            if response_data_key in data:
                return data[response_data_key]
            return None
        else:
            return data

    def regions(self, continent='us', include_gov=False):
        # returns (string, ...)
        regions = self._service.region_names
        if continent and continent != "all":
            regions = [r for r in regions
                       if r.startswith("{}-".format(continent))]
        if not include_gov:
            regions = [r for r in regions if "-gov-" not in r]
        return regions

    @property
    def environment(self):
        return self.aws.environment


class AwsCollection(object):
    """Abstract class for AWS object collections"""
    def get(self, filters=None):
        raise NotImplementedError()

    def create(self, *args, **kwargs):
        raise NotImplementedError()

    def destroy(self, obj_id):
        raise NotImplementedError()



from acky.api import (
    AwsCollection,
    AwsApiClient,
)


class RDSApiClient(AwsApiClient):
    service_name = "rds"


class RDS(RDSApiClient):
    def engine_versions(engine='mysql', default_only=False):
        pass

    @property
    def Instances(self):
        return InstanceCollection(self._aws)

    @property
    def Snapshots(self):
        return SnapshotCollection(self._aws)

    @property
    def EventSubscriptions(self):
        return EventSubscriptionCollection(self._aws)

    @property
    def SecurityGroups(self):
        return SecurityGroupCollection(self._aws)

    @property
    def SecurityGroupRules(self):
        return SecurityGroupRuleCollection(self._aws)


class InstanceCollection(AwsCollection, RDSApiClient):
    def get(self, filters=None):
        # returns (inst, ...)
        # DescribeDBInstances (no native filter)
        raise NotImplementedError()

    def Launcher(self, config=None):
        class _launcher(object):
            def __init__(self, aws, config):
                # Configurable launcher for RDS instances. Create the Launcher
                # (passing an optional dict of its attributes), set its
                # attributes (as described in the CreateDBnstance API docs),
                # then launch()
                self._aws = aws
                self.config = config

            def launch(self, inst_id, password):
                # returns inst_info
                # CreateDBInstance
                raise NotImplementedError()

        if not config:
            config = {}
        return _launcher(self._aws, config)

    def Replicator(self, config=None):
        class _replicator(object):
            def __init__(self, aws, config):
                # Configurable launcher for RDS Read Replicas. Create the
                # Replicator (passing an optional dict of its attributes), set
                # its attributes (as described in the
                # CreateDBInstanceReadReplica API docs), then replicate()
                self._aws = aws
                self.config = config

            def replicate(self, src, dest):
                # returns replica_info
                # CreateDBInstanceReadReplica
                raise NotImplementedError()

        if not config:
            config = {}
        return _replicator(self._aws, config)

    def destroy(self, inst, final_snap=None):
        # returns bool
        # DeleteDBInstance
        raise NotImplementedError()

    def promote(self, replica, retention, window):
        # returns bool
        # PromoteReadReplica
        raise NotImplementedError()

    def reboot(self, inst):
        # returns bool
        # RebootDBInstance
        raise NotImplementedError()


class SnapshotCollection(AwsCollection, RDSApiClient):
    def get(self, filters=None):
        # returns (snap, ...)
        # DescribeDBSnapshots (no native filter)
        raise NotImplementedError()

    def create(self, inst, snap_id):
        # returns snap
        # CreateDBSnapshot
        raise NotImplementedError()

    def copy(self, orig, new):
        # returns snap
        # CopyDBSnapshot
        raise NotImplementedError()

    def destroy(self, snap):
        # returns bool
        # DeleteDBSnapshot
        raise NotImplementedError()


class EventSubscriptionCollection(AwsCollection, RDSApiClient):
    def get(self, filters=None):
        # returns (sub, ...)
        # DescribeEventSubscriptions (no native filter)
        raise NotImplementedError()

    def create(self, name, topic):
        # returns sub
        # CreateEventSubscription
        raise NotImplementedError()

    def destroy(self, sub):
        # returns bool
        # DeleteEventSubscription
        raise NotImplementedError()

    def categories(self, sourcetype=None):
        # returns (cat, ...)
        # DescribeEventCategories
        raise NotImplementedError()


class SecurityGroupCollection(AwsCollection, RDSApiClient):
    def get(self, filters=None):
        # returns (sg_info, ...)
        # DescribeDBSecurityGroups
        raise NotImplementedError()

    def create(self, name, description):
        # returns sg_info
        # CreateDBSecurityGroup
        raise NotImplementedError()

    def destroy(self, sg):
        # returns bool
        # DeleteDBSecurityGroup
        raise NotImplementedError()


class SecurityGroupRuleCollection(AwsCollection, RDSApiClient):
    def get(self, filters=None):
        # returns (sgr_info, ...)
        # DescribeDBSecurityGroups
        raise NotImplementedError()

    def add(self, sg, rules):
        # returns (added_sgr_info, ...)
        # AuthorizeDBSecurityGroupIngress
        raise NotImplementedError()

    def remove(self, sg, rules):
        # returns (removed_sgr_info, ...)
        # RevokeDBSecurityGroupIngress
        raise NotImplementedError()

    def replace(self, sg, rules):
        # returns (removed, added)
        # AuthorizeDBSecurityGroupIngress
        # RevokeDBSecurityGroupIngress
        raise NotImplementedError()



from acky.api import (
    AwsCollection,
    AwsApiClient,
)


class IAMApiClient(AwsApiClient):
    service_name = "iam"


class IAM(IAMApiClient):
    @property
    def Users(self):
        return UserCollection(self._aws)

    @property
    def Groups(self):
        return GroupCollection(self._aws)

    @property
    def Keys(self):
        return KeysCollection(self._aws)


class UserCollection(AwsCollection, IAMApiClient):
    def get(self, filters=None):
        # returns (user, ...)
        # ListUsers, GetUser
        raise NotImplementedError()

    def get_current(self):
        # returns user
        # GetUser
        return self.call("GetUser", response_data_key="User")

    def create(self):
        # returns user
        # CreateUser
        raise NotImplementedError()

    def destroy(self, user):
        # returns bool
        # DeleteUser
        raise NotImplementedError()


class GroupCollection(AwsCollection, IAMApiClient):
    def get(self, filters=None):
        # returns (grp, ...)
        # ListGroups, GetGroup
        raise NotImplementedError()

    def add_user(self):
        # returns bool
        # AddUserToGroup
        raise NotImplementedError()

    def remove_user(self):
        # returns bool
        # RemoveUserFromGroup
        raise NotImplementedError()


class KeysCollection(AwsCollection, IAMApiClient):
    def create(self, username=None):
        # returns (access_key, secret_key)
        # CreateAccessKey
        if not username:
            username = self._aws.username
        creds = self.call('CreateAccessKey', user_name=username,
                          response_data_key='AccessKey')
        if creds:
            return creds['AccessKeyId'], creds['SecretAccessKey']
        return None

    def destroy(self, access_key, username=None):
        # returns bool
        # DeleteAccessKey
        if not username:
            username = self._aws.username
        self.call('DeleteAccessKey', user_name=username,
                  access_key_id=access_key)
        return True
